<!--GitHub watchers/stars overview (disabled for now)
<p>
  <iframe src="https://ghbtns.com/github-btn.html?user=rnag&repo=dataclass-wizard&type=watch&count=true&size=large"
    allowtransparency="true" frameborder="0" scrolling="0" width="200px" height="35px"></iframe>
</p>
-->
<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/docsearch.js@2/dist/cdn/docsearch.min.css" />
<style>
  .algolia-autocomplete {
    width: 100%;
    height: 1.5em
  }

  .algolia-autocomplete a {
    border-bottom: none !important;
  }

  #doc_search {
    width: 100%;
    height: 100%;
  }
</style>
<!--Search functionality (disabled for now)

<input id="doc_search" placeholder="Search the doc" autofocus />
<script type="text/javascript" src="https://cdn.jsdelivr.net/npm/docsearch.js@2/dist/cdn/docsearch.min.js" onload="docsearch({
  apiKey: 'abc123',
  indexName: 'dataclass-wizard',
  inputSelector: '#doc_search',
  debug: false // Set debug to true if you want to inspect the dropdown
})" async></script>
-->

<h1 class="logo"><a href="{{ pathto(master_doc) }}">Dataclass Wizard</a></h1>

<p class="blurb">
    Bring Python <code>dataclasses</code> to life — the <i>wizard</i> way!

<!--    A set of simple, yet elegant <i>wizarding</i> tools for interacting with the-->
<!--    Python <code>dataclasses</code> module.-->
</p>

<h3>Useful Links</h3>
<ul>
  <!-- Quickstart is hidden because the docs are outdated; will need to update later. -->
  <!--<li><a href="https://dataclass.readthedocs.io/en/latest/user/quickstart/">Quickstart</a></li>-->
  <li><a href="{{ pathto('overview') }}">Overview</a></li>
  <li><a href="{{ pathto('installation') }}">Installation</a></li>
  <li><a href="{{ pathto('examples') }}">Examples</a></li>
  <li><a href="{{ pathto('python_compatibility') }}">Py Compatibility</a></li>
  <li><a href="{{ pathto('common_use_cases', 1) }}">Common Use Cases</a></li>
  <li><a href="{{ pathto('advanced_usage', 1) }}">Advanced Usage</a></li>
  <li><a href="{{ pathto('history') }}">Release History</a></li>
  <li><a href="{{ pathto('contributing') }}">Contributors Guide</a></li>

  <p></p>

  <li><a href="{{ pathto('env_magic') }}">Env Magic</a></li>
  <li><a href="{{ pathto('wiz_cli') }}">CLI Tool</a></li>
  <li><a href="{{ pathto('using_field_properties') }}">Using Field Properties</a></li>

  <p></p>

  <li><a href="https://github.com/rnag/dataclass-wizard/discussions">🗣 GitHub Discussions</a></li>
  <li><a href="https://github.com/rnag/dataclass-wizard/issues">🎯 Issue Tracker</a></li>

  <p></p>
  <i>Dataclass Wizard @</i>
  <li>
      <a href="https://github.com/rnag/dataclass-wizard">GitHub</a> |
      <a href="https://pypi.org/project/dataclass-wizard">PyPI</a> |
      <a href="https://anaconda.org/conda-forge/dataclass-wizard">Anaconda</a> |
      <a href="https://stackoverflow.com/questions/tagged/dataclass-wizard">Stack Overflow</a>
  </li>
</ul>
