"""
Dataclass Wizard - a set of wizarding tools for interacting with `dataclasses`
"""

__title__ = 'dataclass-wizard'

__description__ = ('Lightning-fast JSON wizardry for Python dataclasses — '
                   'effortless serialization right out of the box!')
__url__ = 'https://github.com/rnag/dataclass-wizard'
__version__ = '1.0.0'
__author__ = 'Ritvik Nag'
__author_email__ = 'me@ritviknag.com'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2021-2026 Ritvik Nag'
