def to_camel_case(string: str) -> str: ...
def to_pascal_case(string): ...
def to_lisp_case(string: str) -> str: ...
def to_snake_case(string: str) -> str: ...
def replace_multi_with_single(string: str, char: str = ...) -> str: ...
