__all__ = ['normalize',
           'possible_json_keys',
           'possible_env_vars',
           'repl_or_with_union']

from ..enums import EnvKeyStrategy

def normalize(string: str) -> str: ...
def possible_json_keys(field: str) -> list: ...
def possible_env_vars(field: str, lookup_strat: EnvKeyStrategy) -> list: ...
def to_camel_case(string: str) -> str: ...
def to_pascal_case(string): ...
def to_lisp_case(string: str) -> str: ...
def to_snake_case(string: str) -> str: ...
def repl_or_with_union(s: str): ...
