from typing import Any

from .._type_def import FREF

__all__ = ['is_union',
           'get_origin',
           'get_origin_v2',
           'is_typed_dict_type_qualifier',
           'get_args',
           'get_keys_for_typed_dict',
           'is_typed_dict',
           'is_generic',
           'is_annotated',
           'eval_forward_ref',
           'eval_forward_ref_if_needed']

def get_args(tp: Any) -> tuple[Any, ...]: ...
def get_keys_for_typed_dict(cls): ...
def is_typed_dict_type_qualifier(cls) -> bool: ...
def is_union(cls) -> bool: ...
def get_origin_v2(cls): ...
def is_typed_dict(cls: type) -> bool: ...
def is_generic(cls): ...
def get_origin(cls, raise_: bool = ...): ...
def is_annotated(cls): ...
def eval_forward_ref(base_type: FREF, cls: type): ...
def eval_forward_ref_if_needed(base_type, base_cls: type): ...
